;;======================================================================;;
;;			GenHSI	 					;;
;;======================================================================;;
;;									;;
;; Program:		Charlieplexing test	;;
;; Code:		Jindra Ficik					;;
;; Platform:		Microchip PIC16F628, 4 Mhz			;;
;; Date:		14.04.2013					;;
;; First release:	14.04.2013					;;
;; LastDate:		14.04.2013					;;
;;									;;
;;======================================================================;;

; Processor PIC 16F628 running at 4MHz.
;
; Revisions:
;
; 14/04/2013	Start of writing code.


; ----- Definitions

#define		__VERNUM	0x01
#define		__VERREV	"A"
#define		__VERDAY	0x14
#define		__VERMONTH	0x04
#define		__VERYEAR	0x13


                list    p=16F628a,r=hex

		errorlevel -302
		errorlevel -306

	        INCLUDE "P16F628A.INC"

                __FUSES _BODEN_ON & _CP_OFF & _PWRTE_ON & _WDT_OFF & _LVP_OFF & _MCLRE_OFF  & _INTRC_OSC_NOCLKOUT


#define		RAMINI0		0x020		; 80 bytes
#define		RAMINI1		0x0A0		; 80 bytes
#define		RAMINI2		0x120		; 48 bytes
#define		RAMINT		0x070		; 16 bytes

CHPORT		equ		PORTB		; port for charlieplexing
CHTRIS		equ		TRISB		; tri state for charlieplexing

; ----- Macros

SEL_BANK_0:	macro	
		bcf	STATUS,RP0		; macros for data access
		bcf	STATUS,RP1
		endm

SEL_BANK_1:	macro	
		bsf	STATUS,RP0
		bcf	STATUS,RP1
		endm

SEL_BANK_2:	macro	
		bcf	STATUS,RP0
		bsf	STATUS,RP1
		endm

SEL_BANK_3:	macro	
		bsf	STATUS,RP0
		bsf	STATUS,RP1
		endm

#define		SEL_IBANK_0	bcf	STATUS,IRP

#define		SEL_IBANK_1	bsf	STATUS,IRP

#define		DNOP		goto	$+1


; ----- Constant values

FXTAL		equ	D'4000000'		; Crystal frequency


RA_TRIS     	equ     0xFF			; all input
RB_TRIS     	equ     0xFF			; all input
RA_INI      	equ     0x00			; TXD, CTS high
RB_INI      	equ     0x00			; 

INTC_INI	equ	0xC0	; PEIE - Prhepial int enable, global int enable
PIE1_INI	equ	0x02	; TMR2IE - Timer 2 int enable (every 1ms)
T2CON_INI equ	0x00	; Prescaler 1:1, Timer 2 disable, must be enabled on demand
PR2_INI equ	0xF9	; 249 dec = 250 instruction = 0.25ms without prescaler
;PR2_INI equ	0x7C	; 124 dec = 125 instruction = 1/8ms without prescaler
;PR2_INI equ	0x52	; 82 dec = 83 instruction ~ 1/12ms without prescaler
;PR2_INI equ	0x3E	; 62 dec = 63 instruction ~ 1/16ms without prescaler
;T2CON_INI equ	0x01	; Prescaler 1:4, Timer 2 disable, must be enabled on demand
;PR2_INI equ	0xF9	; 249 dec = 250 instruction = 1ms with prescaler
;PR2_INI equ	0x7C	; 124 dec = 125 instruction = 0.5ms with prescaler
;PR2_INI equ	0x3F	; 63 dec = 64 instruction = 1ms with prescaler
OPTION_INI	equ	0x82	; Option reg: timer prescaler 1:4, pull-ups
			; Port B Pull-up:       	1   disabled
			; External int edge:    	0   falling
			; Timer trigger source: 	0   clock
			; Timer trigger edge:   	0   rising
			; Prescaler assignment: 	0   Timer0
			; Prescaler value:	010   1:8


; ----- Variables

; --- Top on all banks Section

INT_W		equ	RAMINT+0x00		; Interrupt variables. Last RAM locations
INT_STAT	equ	RAMINT+0x01
INT_PCLATH	equ	RAMINT+0x02
INT_FSR		equ	RAMINT+0x03

EEDATA0		equ	RAMINT+0x04		; EEPROM shadow variables
EEADR0		equ	RAMINT+0x05

; --- Bank 0

NXT_PORT	equ	RAMINI0+0x00		; Data for next port character
NXT_TRIS	equ	RAMINI0+0x01		; Data for next tristate character
BIT_CNT	equ	RAMINI0+0x02		; Counter of bits (pixels)
MSCNT	equ	RAMINI0+0x03		; Counter of miliseconds (decreased every 1ms)
FLAGS	equ	RAMINI0+0x04		; various flags

MSB		equ	RAMINI0+0x05		; pointer to characters H
LSB		equ	RAMINI0+0x06		; pointer to characters L
STEPS	equ	RAMINI0+0x07		; number of steps left

COUNT	equ	RAMINI0+0x08		; universal counter

SCRINI	equ	RAMINI0+0x30		; First screen buffer location
SCR0  	equ	RAMINI0+0x30		; Screen buffer - row 1
SCR1  	equ	RAMINI0+0x31		; Screen buffer - row 1
SCR2  	equ	RAMINI0+0x32		; Screen buffer - row 1
SCR3  	equ	RAMINI0+0x33		; Screen buffer - row 1
SCR4  	equ	RAMINI0+0x34		; Screen buffer - row 1
SCR5  	equ	RAMINI0+0x35		; Screen buffer - row 1
SCR6  	equ	RAMINI0+0x36		; Screen buffer - row 1
SCREND	equ	RAMINI0+0x36		; Last screen buffer location



; --- Flags

#define		RIGHT_LINE	FLAGS,0		; Reading right line
#define		RCV_S_CMD	FLAGS,1		; Received first 's' command
#define		RXEMPTY		FLAGS,4		; Serial receive buffer empty
#define		TERMINAL_MODE	FLAGS,5		; Terminal mode


; ------ Program Section

		org	0x000

PowerUp:
		clrf	STATUS			; ensure we are at bank 0
		clrf	PCLATH			; ensure page bits before goto !!
		clrf	INTCON			; disable all interrupts
		goto	Start


;
;**********************************************************************************************************************
; ISR (Interrupt Service Routines)
;**********************************************************************************************************************
;
	org	0x004
Interrupt:
	movwf	INT_W	; 1 Save registers W, STATUS, PCLATH, FSR
	swapf	STATUS,w	; 2
	clrf	STATUS	; 3 Ensure port / timer access in bank 0
	movwf	INT_STAT	; 4
	movf	PCLATH,w	; 
	movwf	INT_PCLATH	; save PCLATH, interrupt uses page 0
	clrf	PCLATH	; 
	movf	FSR,w
	movwf	INT_FSR

	bcf	PIR1,TMR2IF	; open int for timer2
	clrf	CHPORT		; off all leds
	comf	NXT_TRIS,w	; prepare tri state
	bsf	STATUS,RP0	; bank 1
	movwf	CHTRIS		; upload tri state
	bcf	STATUS,RP0	; bank 0
	movf	NXT_PORT,w	; prepare port data
	movwf	CHPORT		; upload port data

	movf	BIT_CNT,w	; we are refreshing 4*per ms, then divide it
	andlw	0x03
	btfsc	STATUS,Z
	decf	MSCNT,f


	incf	BIT_CNT,f	; next bit
	movlw	0x38		; last bit is 55d=37h
	; movlw	0x10		; last bit is 16d=10h
	xorwf	BIT_CNT,w	; last bit
	btfsc	STATUS,Z
	clrf	BIT_CNT		; clear counter

	movf	BIT_CNT,w
	call	BitConvert
	movwf	NXT_PORT
	movf	BIT_CNT,w
	call	TrisConvert
	movwf	NXT_TRIS

	rlf		BIT_CNT,w	; BIT_CNT=xxAAADDD (AAA = byte DDD = bit in bite)
	andlw	0x70		; 0AAA0000
	movwf	FSR
	swapf	FSR,w		; 00000AAA
	addlw	SCRINI		; + first addr of buffer
	movwf	FSR			; into FSR
	movf	BIT_CNT,w
	call	BitConvert	; bit position of DDD
	andwf	INDF,w		; check if is 0 or 1
	btfsc	STATUS,Z	; if zero = dark
	clrf	NXT_TRIS	; disable leds
         
EndInt:
	movf	INT_FSR,w
	movwf	FSR
	movf	INT_PCLATH,w
	movwf	PCLATH
	swapf	INT_STAT,w
	movwf	STATUS
	swapf	INT_W,f
	swapf	INT_W,w
	retfie


; ----- Tables

BitConvert:				; also for conversion of portb mask
		andlw	0x07
		addwf	PCL,f		; Convert bit position to byte mask
		retlw	0x01
		retlw	0x02
		retlw	0x04
		retlw	0x08
		retlw	0x10
		retlw	0x20
		retlw	0x40
		retlw	0x80

RBitConvert:				
		andlw	0x07
		addwf	PCL,f		; Convert bit position to byte mask - in reverse
		retlw	0x80
		retlw	0x40
		retlw	0x20
		retlw	0x10
		retlw	0x08
		retlw	0x04
		retlw	0x02
		retlw	0x01

TrisConvert:
		addwf	PCL,f		; Convert position to tris mask
		retlw 0x03		;0
		retlw 0x03		;1
		retlw 0x05
		retlw 0x09
		retlw 0x11
		retlw 0x21
		retlw 0x41
		retlw 0x81
		retlw 0x05
		retlw 0x06
		retlw 0x06
		retlw 0x0A
		retlw 0x12
		retlw 0x22
		retlw 0x42
		retlw 0x82
		retlw 0x09
		retlw 0x0A
		retlw 0x0C
		retlw 0x0C
		retlw 0x14
		retlw 0x24
		retlw 0x44
		retlw 0x84
		retlw 0x11
		retlw 0x12
		retlw 0x14
		retlw 0x18
		retlw 0x18
		retlw 0x28
		retlw 0x48
		retlw 0x88
		retlw 0x21
		retlw 0x22
		retlw 0x24
		retlw 0x28
		retlw 0x30
		retlw 0x30
		retlw 0x50
		retlw 0x90
		retlw 0x41
		retlw 0x42
		retlw 0x44
		retlw 0x48
		retlw 0x50
		retlw 0x60
		retlw 0x60
		retlw 0xA0
		retlw 0x81
		retlw 0x82
		retlw 0x84
		retlw 0x88
		retlw 0x90
		retlw 0xA0
		retlw 0xC0		; 54
		retlw 0xC0		; 55
                 
        if ($ > d'255')
                ERROR "  Tables exceded page 0.   If it works, why do you change it?   "
        endif


; ----- Power Up Section ( Be sure is in first 2K program code)

Start:
		movlw	0x07			; PIC 16F628, all digital
		movwf	CMCON                          
		clrf	RCSTA			; disable UART

		movlw 	RA_INI			; Set ports
		movwf 	PORTA                         
		movlw 	RB_INI
		movwf 	PORTB
		SEL_BANK_1
		movlw	PIE1_INI
		movwf	PIE1			; peripheral interrupts
		movlw	RB_TRIS
		movwf 	TRISB
		movlw 	RA_TRIS         	; Set port A I/O configuration
		movwf 	TRISA
		movlw 	OPTION_INI      	; PORTB pull-ups disabled, Timer configuration
		movwf 	OPTION_REG
		movlw	0x0B			; internal oscillator to 4MHz
		movwf	PCON
		movlw	0x00	; VRR off voltage reference
		movwf	VRCON
		movlw	PR2_INI
		movwf	PR2

		SEL_BANK_0

		clrf	TMR0			; init timers
		movlw	T2CON_INI
		movwf	T2CON
		movlw	0x07	; RA0, RA1, RA2 digital inputs
		movwf	CMCON                          

		movlw	RAMINI0			; Clear variables bank 0 & bank 1
		movwf	FSR
ClearRAM:
		clrf	INDF
		incf	FSR,f
		btfss	FSR,7	
		goto	ClearRAM

		movlw	INTC_INI		; Set interrupts
		movwf	INTCON

		bcf	PIR1,TMR2IF	; open int for timer2
		bsf	INTCON,GIE	; enable interupt
		bsf	T2CON,TMR2ON	; enable timer2

		movlw b'01110000'
		movwf SCR0
		movlw b'00100000'
		movwf SCR1
		movlw b'10100000'
		movwf SCR2
		movlw b'01001110'
		movwf SCR3
		movlw b'00001000'
		movwf SCR4
		movlw b'00001100'
		movwf SCR5
		movlw b'00001001'
		movwf SCR6

		movlw d'250'
		movwf MSCNT

		movlw HIGH(HeloWorld)
		movwf MSB
		movlw LOW(HeloWorld)
		movwf LSB

StartLoop1:
		;goto	MainLoop
		movf	MSCNT,w
		btfss	STATUS,Z
		goto	StartLoop1	; wait until next timestamp

		movlw d'250'
		movwf MSCNT

StartLoop2:
		;goto	MainLoop
		movf	MSCNT,w
		btfss	STATUS,Z
		goto	StartLoop2	; wait until next timestamp

		movlw d'250'
		movwf MSCNT

StartLoop3:
		;goto	MainLoop
		movf	MSCNT,w
		btfss	STATUS,Z
		goto	StartLoop3	; wait until next timestamp

		movlw d'250'
		movwf MSCNT


; --- Main Loop
MainLoop:
		;goto	MainLoop
		movf	MSCNT,w
		btfss	STATUS,Z
		goto	MainLoop	; wait until next timestamp

		movlw d'110'
		movwf MSCNT

		movlw	SCREND
		movwf	FSR
		movlw	6
		movwf	COUNT
ShiftCols:
		decf	FSR,f
		movf	INDF,w
		incf	FSR,f
		movwf	INDF
		decf	FSR,f
		decfsz	COUNT,f
		goto	ShiftCols
		call	GetLine
		clrf	PCLATH
		movwf	INDF
        incf	LSB,f        ; next line
        btfsc	STATUS,Z
        incf	MSB,f
		movlw 	LOW(EndHW)
		xorwf	LSB,w
		btfss	STATUS,Z
		goto	MainLoop
		movlw	 HIGH(EndHW)
		xorwf	MSB,w
		btfss	STATUS,Z
		goto	MainLoop

		movlw HIGH(HeloWorld)
		movwf MSB
		movlw LOW(HeloWorld)
		movwf LSB
		goto  MainLoop

GetLine:
		movf	MSB,w
		movwf	PCLATH
		movf	LSB,w
		movwf	PCL


Delay10us:
		DNOP				; 1,2	Delay w * 10us
		DNOP				; 3,4
		DNOP				; 5,6

		addlw	0xFF			; 7
		btfss	STATUS,Z		; 8
		goto	Delay10us		; 9,10
		return

HeloWorld:
	retlw b'11000011' ;1
	retlw b'11111111' ;2
	retlw b'11011011' ;3
	retlw b'00011000' ;4
	retlw b'11011011' ;5
	retlw b'11111111' ;6
	retlw b'11000011' ;7
	retlw b'00000000' ;8
	retlw b'11000011' ;9
	retlw b'11111111' ;10
	retlw b'11011011' ;11
	retlw b'11011011' ;11
	retlw b'11000011' ;12
	retlw b'00000000' ;13
	retlw b'11000011' ;14
	retlw b'11111111' ;15
	retlw b'11000011' ;16
	retlw b'11000000' ;17
	retlw b'11000000' ;17
	retlw b'00000000' ;18
	retlw b'00111100' ;19
	retlw b'11100111' ;20
	retlw b'11000011' ;21
	retlw b'11100111' ;22
	retlw b'00111100' ;23
	retlw b'00000000' ;24
	retlw b'00000000' ;24
	retlw b'00000000' ;24
	retlw b'00000000' ;25
	retlw b'00111111' ;26
	retlw b'11100000' ;27
	retlw b'00110000' ;28
	retlw b'11100000' ;29
	retlw b'00011111' ;30
	retlw b'00000000' ;31
	retlw b'01111000' ;32
	retlw b'10000100' ;33
	retlw b'10000100' ;34
	retlw b'01111000' ;35
	retlw b'00000000' ;36
	retlw b'11111100' ;37
	retlw b'00011000' ;38
	retlw b'00001100' ;39
	retlw b'00000000' ;40
	retlw b'11111110' ;41
	retlw b'00000000' ;42
	retlw b'01111000' ;43
	retlw b'10000100' ;44
	retlw b'01001000' ;45
	retlw b'11111111' ;46
	retlw b'00000000' ;47
	retlw b'00000000' ;48
	retlw b'00000000' ;49
	retlw b'00000000' ;50
	retlw b'00000000' ;51
	retlw b'00000000' ;52
EndHW:
	retlw b'00000000' ;53


; ---------------------- Internal EEPROM -------------------------------------

		org	0x2100

		dt	"CharliPX"
		dt	" J.Fucik"
		dt	(__VERDAY   >> 4)  +0x30
		dt	(__VERDAY   & 0x0F)+0x30,"/"
		dt	(__VERMONTH >> 4)  +0x30
		dt	(__VERMONTH & 0x0F)+0x30,"/"
		dt	(__VERYEAR  >> 4)  +0x30
		dt	(__VERYEAR  & 0x0F)+0x30

		end